MANIFEST_BUILD: str = "1.20.80"  # The build version of the manifest.
BLOCK_SERVER_VERSION: str = "1.21.20"  # The version of the block server.
ENTITY_SERVER_VERSION: str = "1.21.51"  # The version of the entity server.
ENTITY_CLIENT_VERSION: str = "1.10.0"  # The version of the entity client.
BP_ANIMATION_VERSION: str = "1.10.0"  # The version of the behavior pack animation.
RP_ANIMATION_VERSION: str = "1.8.0"  # The version of the resource pack animation.
ANIMATION_CONTROLLERS_VERSION: str = "1.10.0"  # The version of the animation controllers.
SPAWN_RULES_VERSION: str = "1.8.0"  # The version of the spawn rules.
GEOMETRY_VERSION: str = "1.21.20"  # The version of the geometry.
RENDER_CONTROLLER_VERSION: str = "1.10.0"  # The version of the render controller.
SOUND_DEFINITIONS_VERSION: str = "1.20.20"  # The version of the sound definitions.
DIALOGUE_VERSION: str = "1.18.0"  # The version of the dialogue.
FOG_VERSION: str = "1.16.100"  # The version of the fog.
MATERIALS_VERSION: str = "1.0.0"  # The version of the materials.
ITEM_SERVER_VERSION: str = "1.21.30"  # The version of the item server.
GLOBAL_LIGHTING = "1.0.0"
CAMERA_PRESET_VERSION = "1.21.0"  # The version of the camera presets.
BLOCK_JSON_FORMAT_VERSION = "1.1.0"  # The version of the block json format.

MODULE_MINECRAFT_SERVER: str = "1.16.0"  # The version of the Minecraft server module.
MODULE_MINECRAFT_SERVER_UI: str = "1.3.0"  # The version of the Minecraft UI module.
MODULE_MINECRAFT_SERVER_EDITOR: str = "0.1.0"  # The version of the Minecraft UI module.
MODULE_MINECRAFT_SERVER_GAMETEST: str = "1.0.0"  # The version of the Minecraft UI module.
